﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Media3D;

namespace TestProject
{
    /// <summary>
    /// 메인 윈도우
    /// </summary>
    public partial class MainWindow : Window
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 3차원 모델 그룹
        /// </summary>
        private Model3DGroup model3Dgroup = new Model3DGroup();

        /// <summary>
        /// 카메라
        /// </summary>
        private PerspectiveCamera camera;

        /// <summary>
        /// 카메라 파이
        /// </summary>
        private double cameraPhi = Math.PI / 6.0;

        /// <summary>
        /// 카메라 세타
        /// </summary>
        private double cameraTheta = Math.PI / 6.0;

        /// <summary>
        /// 카메라 R
        /// </summary>
        private double cameraR = 4.0;

        /// <summary>
        /// 카메라 델타 파이
        /// </summary>
        private const double CAMERA_DELTA_PHI = 0.1;

        /// <summary>
        /// 카메라 델타 세타
        /// </summary>
        private const double CAMERA_DELTA_THETA = 0.1;

        /// <summary>
        /// 카메라 델타 R
        /// </summary>
        private const double CAMERA_DELTA_R = 0.1;

        /// <summary>
        /// 3차원 모델 딕셔너리
        /// </summary>
        private Dictionary<Model3D, string> model3dDictionary = new Dictionary<Model3D, string>();

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MainWindow()

        /// <summary>
        /// 생성자
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 윈도우 로드시 처리하기 - Window_Loaded(sender, e)

        /// <summary>
        /// 윈도우 로드시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            this.camera = new PerspectiveCamera();

            this.camera.FieldOfView = 60;

            this.viewport3D.Camera = this.camera;

            SetCameraPosition();

            DefineLight();

            DefineModel(this.model3Dgroup);

            ModelVisual3D modelVisual3D = new ModelVisual3D();

            modelVisual3D.Content = this.model3Dgroup;

            this.viewport3D.Children.Add(modelVisual3D);
        }

        #endregion
        #region 윈도우 키 DOWN 처리하기 - Window_KeyDown(sender, e)

        /// <summary>
        /// 윈도우 키 DOWN 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Window_KeyDown(object sender, KeyEventArgs e)
        {
            switch(e.Key)
            {
                case Key.Up :

                    this.cameraPhi += CAMERA_DELTA_PHI;

                    if(this.cameraPhi > Math.PI / 2.0)
                    {
                        this.cameraPhi = Math.PI / 2.0;
                    }

                    break;

                case Key.Down :

                    this.cameraPhi -= CAMERA_DELTA_PHI;

                    if(this.cameraPhi < -Math.PI / 2.0)
                    {
                        this.cameraPhi = -Math.PI / 2.0;
                    }

                    break;

                case Key.Left :

                    this.cameraTheta += CAMERA_DELTA_THETA;

                    break;

                case Key.Right :

                    this.cameraTheta -= CAMERA_DELTA_THETA;

                    break;

                case Key.Add     :
                case Key.OemPlus :

                    this.cameraR -= CAMERA_DELTA_R;

                    if(this.cameraR < CAMERA_DELTA_R)
                    {
                        this.cameraR = CAMERA_DELTA_R;
                    }

                    break;

                case Key.Subtract :
                case Key.OemMinus :

                    this.cameraR += CAMERA_DELTA_R;

                    break;
            }

            SetCameraPosition();
        }

        #endregion
        #region 3차원 뷰포트 마우스 DOWN 처리하기 - viewport3D_MouseDown(sender, e)

        /// <summary>
        /// 3차원 뷰포트 마우스 DOWN 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void viewport3D_MouseDown(object sender, MouseButtonEventArgs e)
        {
            Point mousePoint = e.GetPosition(this.viewport3D);

            HitTestResult hitTestResult = VisualTreeHelper.HitTest(this.viewport3D, mousePoint);

            RayMeshGeometry3DHitTestResult result = hitTestResult as RayMeshGeometry3DHitTestResult;

            if(result == null)
            {
                Title = "3차원 모델 히트 테스트 하기";
            }
            else
            {
                Title = string.Format("3차원 모델 히트 테스트 하기 : {0}", this.model3dDictionary[result.ModelHit]);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region 카메라 위치 설정하기 - SetCameraPosition()

        /// <summary>
        /// 카메라 위치 설정하기
        /// </summary>
        private void SetCameraPosition()
        {
            double y   = this.cameraR * Math.Sin(this.cameraPhi);
            double hyp = this.cameraR * Math.Cos(this.cameraPhi);
            double x   = hyp * Math.Cos(this.cameraTheta);
            double z   = hyp * Math.Sin(this.cameraTheta);

            this.camera.Position = new Point3D(x, y, z);

            this.camera.LookDirection = new Vector3D(-x, -y, -z);

            this.camera.UpDirection = new Vector3D(0, 1, 0);
        }

        #endregion
        #region 조명 정의하기 - DefineLight()

        /// <summary>
        /// 조명 정의하기
        /// </summary>
        private void DefineLight()
        {
            AmbientLight ambientLight = new AmbientLight(Colors.Gray);

            DirectionalLight directionalLight = new DirectionalLight(Colors.Gray, new Vector3D(-1.0, -3.0, -2.0));

            this.model3Dgroup.Children.Add(ambientLight);
            this.model3Dgroup.Children.Add(directionalLight);
        }

        #endregion
        #region 삼각형 추가하기 - AddTriangle(meshGeometry3D, point3D1, point3D2, point3D3)

        /// <summary>
        /// 삼각형 추가하기
        /// </summary>
        /// <param name="meshGeometry3D">3차원 메쉬 기하</param>
        /// <param name="point3D1">3차원 포인트 1</param>
        /// <param name="point3D2">3차원 포인트 2</param>
        /// <param name="point3D3">3차원 포인트 3</param>
        private void AddTriangle(MeshGeometry3D meshGeometry3D, Point3D point3D1, Point3D point3D2, Point3D point3D3)
        {
            int index = meshGeometry3D.Positions.Count;

            meshGeometry3D.Positions.Add(point3D1);
            meshGeometry3D.Positions.Add(point3D2);
            meshGeometry3D.Positions.Add(point3D3);

            meshGeometry3D.TriangleIndices.Add(index++);
            meshGeometry3D.TriangleIndices.Add(index++);
            meshGeometry3D.TriangleIndices.Add(index  );
        }

        #endregion
        #region 스케일 벡터 구하기 - GetScaleVector(vector3D, length)

        /// <summary>
        /// 스케일 벡터 구하기
        /// </summary>
        /// <param name="vector3D">3차원 벡터</param>
        /// <param name="length">길이</param>
        /// <returns>스케일 벡터</returns>
        private Vector3D GetScaleVector(Vector3D vector3D, double length)
        {
            double scale = length / vector3D.Length;

            return new Vector3D
            (
                vector3D.X * scale,
                vector3D.Y * scale,
                vector3D.Z * scale
            );
        }

        #endregion
        #region 세그먼트 추가하기 - AddSegment(meshGeometry3D, point3D1, point3D2, upVector3D, thickness, extend)

        /// <summary>
        /// 세그먼트 추가하기
        /// </summary>
        /// <param name="meshGeometry3D">3차원 메쉬 기하</param>
        /// <param name="point3D1">3차원 포인트 1</param>
        /// <param name="point3D2">3차원 포인트 2</param>
        /// <param name="upVector3D">UP 3차원 벡터</param>
        /// <param name="thickness">두께</param>
        /// <param name="extend">확장 여부</param>
        public void AddSegment(MeshGeometry3D meshGeometry3D, Point3D point3D1, Point3D point3D2, Vector3D upVector3D, double thickness, bool extend)
        {
            Vector3D vector = point3D2 - point3D1;

            if(extend)
            {
                Vector3D scaleVector = GetScaleVector(vector, thickness / 2.0);

                point3D1 -= scaleVector;
                point3D2 += scaleVector;
            }

            Vector3D scaleVector1 = GetScaleVector(upVector3D, thickness / 2.0);
            Vector3D scaleVector2 = Vector3D.CrossProduct(vector, scaleVector1);

            scaleVector2 = GetScaleVector(scaleVector2, thickness / 2.0);

            Point3D p1pp = point3D1 + scaleVector1 + scaleVector2;
            Point3D p1mp = point3D1 - scaleVector1 + scaleVector2;
            Point3D p1pm = point3D1 + scaleVector1 - scaleVector2;
            Point3D p1mm = point3D1 - scaleVector1 - scaleVector2;
            Point3D p2pp = point3D2 + scaleVector1 + scaleVector2;
            Point3D p2mp = point3D2 - scaleVector1 + scaleVector2;
            Point3D p2pm = point3D2 + scaleVector1 - scaleVector2;
            Point3D p2mm = point3D2 - scaleVector1 - scaleVector2;

            AddTriangle(meshGeometry3D, p1pp, p1mp, p2mp);
            AddTriangle(meshGeometry3D, p1pp, p2mp, p2pp);

            AddTriangle(meshGeometry3D, p1pp, p2pp, p2pm);
            AddTriangle(meshGeometry3D, p1pp, p2pm, p1pm);

            AddTriangle(meshGeometry3D, p1pm, p2pm, p2mm);
            AddTriangle(meshGeometry3D, p1pm, p2mm, p1mm);

            AddTriangle(meshGeometry3D, p1mm, p2mm, p2mp);
            AddTriangle(meshGeometry3D, p1mm, p2mp, p1mp);

            AddTriangle(meshGeometry3D, p1pp, p1pm, p1mm);
            AddTriangle(meshGeometry3D, p1pp, p1mm, p1mp);

            AddTriangle(meshGeometry3D, p2pp, p2mp, p2mm);
            AddTriangle(meshGeometry3D, p2pp, p2mm, p2pm);
        }

        #endregion
        #region 케이지 추가하기 - AddCage(meshGeometry3D)

        /// <summary>
        /// 케이지 추가하기
        /// </summary>
        /// <param name="meshGeometry3D">3차원 메쉬 기하</param>
        private void AddCage(MeshGeometry3D meshGeometry3D)
        {
            Vector3D upVector3D = new Vector3D(0, 1, 0);

            AddSegment(meshGeometry3D, new Point3D( 1, 1,  1), new Point3D( 1, 1, -1), upVector3D, 0.05, true);
            AddSegment(meshGeometry3D, new Point3D( 1, 1, -1), new Point3D(-1, 1, -1), upVector3D, 0.05, true);
            AddSegment(meshGeometry3D, new Point3D(-1, 1, -1), new Point3D(-1, 1,  1), upVector3D, 0.05, true);
            AddSegment(meshGeometry3D, new Point3D(-1, 1,  1), new Point3D( 1, 1,  1), upVector3D, 0.05, true);

            AddSegment(meshGeometry3D, new Point3D( 1, -1,  1), new Point3D( 1, -1, -1), upVector3D, 0.05, true);
            AddSegment(meshGeometry3D, new Point3D( 1, -1, -1), new Point3D(-1, -1, -1), upVector3D, 0.05, true);
            AddSegment(meshGeometry3D, new Point3D(-1, -1, -1), new Point3D(-1, -1,  1), upVector3D, 0.05, true);
            AddSegment(meshGeometry3D, new Point3D(-1, -1,  1), new Point3D( 1, -1,  1), upVector3D, 0.05, true);

            Vector3D rightVector3D = new Vector3D(1, 0, 0);

            AddSegment(meshGeometry3D, new Point3D( 1, -1,  1), new Point3D( 1, 1,  1), rightVector3D, 0.05, true);
            AddSegment(meshGeometry3D, new Point3D( 1, -1, -1), new Point3D( 1, 1, -1), rightVector3D, 0.05, true);
            AddSegment(meshGeometry3D, new Point3D(-1, -1,  1), new Point3D(-1, 1,  1), rightVector3D, 0.05, true);
            AddSegment(meshGeometry3D, new Point3D(-1, -1, -1), new Point3D(-1, 1, -1), rightVector3D, 0.05, true);
        }

        #endregion
        #region 모델 정의하기 - DefineModel(model3DGroup)

        /// <summary>
        /// 모델 정의하기
        /// </summary>
        /// <param name="model3DGroup">3차원 모델 그룹</param>
        private void DefineModel(Model3DGroup model3DGroup)
        {
            MeshGeometry3D meshGeometry3D1 = new MeshGeometry3D();

            AddTriangle
            (
                meshGeometry3D1,
                new Point3D(1, 1, 1),
                new Point3D(-1, -1, 1),
                new Point3D(1, -1, -1)
            );

            AddTriangle
            (
                meshGeometry3D1,
                new Point3D(1, 1, 1),
                new Point3D(-1, 1, -1),
                new Point3D(-1, -1, 1)
            );

            AddTriangle
            (
                meshGeometry3D1,
                new Point3D(1, 1, 1),
                new Point3D(1, -1, -1),
                new Point3D(-1, 1, -1)
            );

            AddTriangle
            (
                meshGeometry3D1,
                new Point3D(-1, -1, 1),
                new Point3D(-1, 1, -1),
                new Point3D(1, -1, -1)
            );

            SolidColorBrush solidColorBrush1 = Brushes.Green;
            DiffuseMaterial diffuseMaterial1 = new DiffuseMaterial(solidColorBrush1);
            GeometryModel3D geometryModel3D1 = new GeometryModel3D(meshGeometry3D1, diffuseMaterial1);

            model3DGroup.Children.Add(geometryModel3D1);

            this.model3dDictionary.Add(geometryModel3D1, "녹색 모델");

            MeshGeometry3D meshGeometry3D2 = new MeshGeometry3D();

            AddTriangle
            (
                meshGeometry3D2,
                new Point3D(-1, -1, -1),
                new Point3D(-1, 1, 1),
                new Point3D(1, 1, -1)
            );

            AddTriangle
            (
                meshGeometry3D2,
                new Point3D(-1, -1, -1),
                new Point3D(1, -1, 1),
                new Point3D(-1, 1, 1)
            );

            AddTriangle
            (
                meshGeometry3D2,
                new Point3D(-1, -1, -1),
                new Point3D(1, 1, -1),
                new Point3D(1, -1, 1)
            );

            AddTriangle
            (
                meshGeometry3D2,
                new Point3D(1, -1, 1),
                new Point3D(1, 1, -1),
                new Point3D(-1, 1, 1)
            );

            SolidColorBrush solidColorBrush2 = Brushes.Blue;
            DiffuseMaterial diffuseMaterial2 = new DiffuseMaterial(solidColorBrush2);
            GeometryModel3D geometryModel3D2 = new GeometryModel3D(meshGeometry3D2, diffuseMaterial2);

            model3DGroup.Children.Add(geometryModel3D2);

            this.model3dDictionary.Add(geometryModel3D2, "파랑색 모델");

            MeshGeometry3D meshGeometry3D3 = new MeshGeometry3D();

            AddCage(meshGeometry3D3);

            SolidColorBrush solidColorBrush3 = Brushes.Red;
            DiffuseMaterial diffuseMaterial3 = new DiffuseMaterial(solidColorBrush3);
            GeometryModel3D geometryModel3D3 = new GeometryModel3D(meshGeometry3D3, diffuseMaterial3);

            model3DGroup.Children.Add(geometryModel3D3);

            this.model3dDictionary.Add(geometryModel3D3, "케이지");
        }

        #endregion
    }
}